local name = ...
local mod = LibStub("AceAddon-3.0"):NewAddon("BigBrother_Ready", "AceEvent-3.0", "AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale(name)

mod.version = GetAddOnMetadata(name, 'version')


--[===[@debug@
_G.BBReady = mod

local PREFIX = "|cff00ff00BB_Ready|r: "
local DevTools_Dump
DevTools_Dump = function(value)
	if not _G.DevTools_Dump then
		LoadAddOn("Blizzard_DebugTools")
	end
	DevTools_Dump = _G.DevTools_Dump
	DevTools_Dump(value)
end

local function dump(value)
	local orig = DEFAULT_CHAT_FRAME
	DEFAULT_CHAT_FRAME = ChatFrame4
	DevTools_Dump(value)
	DEFAULT_CHAT_FRAME = orig
end

local function mystrjoin(arg1, ...)
	local text = tostring(arg1)
	for i = 1, select('#', ...) do
		text = text .. tostring(select(i, ...))
	end
	return text
end

local function doDebug(arg1, ...)
	if not arg1 then return end
	if type(arg1) == 'table' then
		ChatFrame4:AddMessage(PREFIX .. "+++ table-dump")
		dump(arg1)
		ChatFrame4:AddMessage(PREFIX .. "--- end of table-dump")
		doDebug(...)
	else
		local text = mystrjoin(PREFIX, arg1, ...)
		ChatFrame4:AddMessage(text)
	end
end
--@end-debug@]===]

local function debug(...)
	--[===[@debug@
	doDebug(...)
	--@end-debug@]===]
end
mod.debug = debug

local foods = {
	[250] = {
		104267, -- 250 Strength
		104273, -- 250 Agility
		104264, -- 250 Intellect
		104278, -- 250 Spirit
		104281, -- 375 Stamina
		124151, -- 250 Haste
		125102, -- 250 Hit
		125104, -- 250 Hit
	},
	[275] = {
		-- The good ones
		104271, -- 275 Strength
		104274, -- 275 Agility
		104276, -- 275 Intellect
		104279, -- 275 Spirit
		104282, -- 415 Stamina
		125106, -- 275 Hit
		125108, -- 275 Expertise
	},
	[300] = {
		-- The really good ones
		104272, -- 300 Strength
		104275, -- 300 Agility
		104277, -- 300 Intellect
		104280, -- 300 Spirit
		104283, -- 450 Stamina
		125113, -- 300 Hit
		125115, -- 300 Expertise
	},
}

local legacyFlasks = {
	-- Flask WotLK
	53752, -- 53752 Lesser Flask of Toughness
	53755, -- 53755 Flask of the Frost Wyrm
	53758, -- 53758 Flask of Stoneblood
	53760, -- 53760 Flask of Pure Mojo
	54212, -- 54212 Flask of Endless Rage
	62380, -- Lesser Flask of Resist
}

local flasks = {
	105689, -- Flask of Spring Blossoms
	105691, -- Flask of the Warm Sun
	105693, -- Flask of Falling Leaves
	105694, -- Flask of the Earth
	105696, -- Flask of Winter\'s Bite
	127230, -- Crystal of Insanity
}

local guardianElixirs = {
	105681, -- Mantid Elixir
	105687, -- Elixir of Mirrors
}

local battleElixirs = {
	105682, -- Mad Hozen Elixir
	105683, -- Elixir of Weaponry
	105684, -- Elixir of the Rapids
	105685, -- Elixir of Peace
	105686, -- Elixir of Perfection
	105688, -- Monk\'s Elixir
}

local db
local timers = {}

local function checkForAnyBuff(unit, buffList)
	for i,v in ipairs(buffList) do
		local spellName = GetSpellInfo(v)
		local name, _, _, _, _, duration, expirationTime,_,_,_,spellid = UnitAura(unit, spellName)
		if name and spellid == v then
			if expirationTime > 0 then
				local remaining = expirationTime - GetTime()
				return remaining
			else
				return 999999
			end
		end
	end
	return
end

local function checkForMissingElixir(unit)
	if db.onlyflasks then
		return L["No flask"]
	end
	local MIN_DURATION = db.threshold
	local guardianRemaining = checkForAnyBuff(unit, guardianElixirs)
	local battleRemaining = checkForAnyBuff(unit, battleElixirs)
	if guardianRemaining and battleRemaining then
		if guardianRemaining > MIN_DURATION and battleRemaining > MIN_DURATION then
			return
		elseif guardianRemaining < MIN_DURATION or battleRemaining < MIN_DURATION then
			return L["Expiring elixirs"]
		end
	end
	local onlybuff = guardianRemaining or battleRemaining
	if onlybuff then
		if onlybuff < MIN_DURATION then
			return L["Only one expiring elixir"]
		else
			return L["Only one elixir"]
		end
	else
		return L["No flask or elixir"]
	end
	return false
end

local function checkForMissingFlask(unit)
	local remaining = checkForAnyBuff(unit, flasks)
	if not remaining then
		remaining = checkForAnyBuff(unit, legacyFlasks)
		if remaining then
			return L["Legacy Flask"]
		else
			return checkForMissingElixir(unit)
		end
	end

	if remaining < db.threshold then
		return L["Expiring flask"], remaining
	else
		return
	end
end

local stillEatingUnits = {}

local WELL_FED = GetSpellInfo(foods[250][1])
local EATING = GetSpellInfo(104235)
local function checkForMissingFood(unit)
	if not db.checkfood then
		return
	end
	local remaining
	for k,v in pairs(foods) do
		if k >= db.foodlevel then
			remaining = checkForAnyBuff(unit, v)
			if remaining then
				break
			end
		end
	end

	if remaining then
		if remaining < db.threshold then
			return L["Expiring food buff"], remaining
		else
			return
		end
	elseif UnitAura(unit, WELL_FED) then
		local _, _, _, _, _, _, _, _, _, _, spellId = UnitAura(unit, WELL_FED)
		local link = GetSpellLink(spellId)
		return L["Low food buff"], link
	elseif UnitAura(unit, EATING) then
		stillEatingUnits[unit] = true
		mod:RegisterEvent("UNIT_AURA")
		return L["Still eating"]
	else
		return L["No food buff"]
	end
end

local function sendMessageToChannel(message, channel)
   if type(channel) == 'number' then
      SendChatMessage(message, "CHANNEL", nil, channel)
   else
      SendChatMessage(message, channel)
   end
end

local function splitStringInTwo(str, maxlength)
   local maxHalf = string.sub(str, 1, maxlength)
   local index = maxHalf:match('^.*(),')
   local half1 = string.sub(str, 1, index-1)
   local half2 = string.sub(str, index+1, #str+1)
   return half1, half2
end

-- split the message into portions of 255 characters because the client will truncate longer messages
local function splitChatMessage(prefix, message, channel)
   local maxlength = 255 - #prefix
   if #message > maxlength then
      local half1, half2 = splitStringInTwo(message, maxlength)
      sendMessageToChannel(prefix .. half1, channel)
      return splitChatMessage(prefix, half2, channel)
   else
      sendMessageToChannel(prefix .. message, channel)
   end
end

local function broadCastToRaid(prefix, table)
	if #table <= 0 then
		return
	end
	local message = prefix .. ": " .. strjoin(", ", unpack(table))
	if db.channel == "SELF" then
		print("BBReady: ", message)
	elseif db.channel == "CUSTOM" then
		-- NYI
	else
		splitChatMessage(prefix .. ": ", (", "):join(unpack(table)), db.channel)
	end
end

function mod:UNIT_AURA(_, unit)
	if not stillEatingUnits[unit] then return end
	if UnitAura(unit, GetSpellInfo(104235)) then return end -- still eating
	stillEatingUnits[unit] = nil
	local state, suffix = checkForMissingFood(unit)
	local name = UnitName(unit)
	if state then
		if suffix then
			broadCastToRaid(state, {("%s (%s)"):format(name, suffix)})
		else
			broadCastToRaid(state, {name})
		end
	end
	if not next(stillEatingUnits) then
		self:UnregisterEvent("UNIT_AURA")
	end
end

local TOKEN = "<BB_Ready>"
local outgoingFilter = function(self, event, msg, ...)
	if msg:find(TOKEN) then return true end
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", outgoingFilter)

local function broadCastWhisper(state, table)
	for _, v in ipairs(table) do
		local name = v:gsub(" \(.*\)", "")
		debug("whispering", name)
		SendChatMessage(("%s %s"):format(TOKEN, state), "WHISPER", nil, name)
	end
end

local function autoInitWithEmptyTable(self, key)
	local val = rawget(self, key)
	if not val then
		val = {}
		rawset(self, key, val)
	end
	return val
end
local mt = { __index = autoInitWithEmptyTable, }

local GROUPS_10_MAN = { true, true }
local GROUPS_25_MAN = { true, true, true, true, true, }
local GROUPS_NONE = {}

local function getRelevantGroups()
	if db.groups.auto then
		local difficulty = GetRaidDifficultyID()
		if difficulty == 3 or difficulty == 5 then -- 10-man
			debug("10-man")
			return GROUPS_10_MAN
		elseif difficulty == 4 or difficulty == 6 then -- 25-man
			debug("25-man")
			return GROUPS_25_MAN
		else
			debug("using no groups because difficulty is ", difficulty)
			return GROUPS_NONE
		end
	else
		return db.groups
	end
end

local function formatTime(seconds)
	if type(seconds) ~= 'number' then
		return seconds
	end
	local minutes = math.floor(seconds/60)
	local seconds = math.floor(seconds - minutes * 60)
	if seconds < 10 then
		seconds = "0" .. seconds
	end
	return ("%s:%s"):format(minutes, seconds)
end

local function iterateRaidMembers()
	local i = 0
	local n = GetNumGroupMembers()
	return function()
			i = i + 1
			if i <= n then
				local unit = "raid" .. i
				local name = UnitName(unit)
				local _, _, group = GetRaidRosterInfo(i)
				return unit, name, group
			end
		end
end

local buffStates = setmetatable({}, mt)
local function handleState(unit, state, remaining)
	if not state then return end
	local name = UnitName(unit)
	if remaining then
		name = ("%s (%s)"):format(name, formatTime(remaining))
	end
	table.insert(buffStates[state], name)
end

function mod:READY_CHECK()
	debug("ready_check")
	wipe(buffStates)
	local groups = getRelevantGroups()
	for unit, name, group in iterateRaidMembers() do
		if not groups[group] then
			debug("skipping ", name, " because player is in group ", group)
			-- nothing
		else
			handleState(unit, checkForMissingFlask(unit))
			handleState(unit, checkForMissingFood(unit))
			if not UnitIsVisible(unit) then
				table.insert(buffStates[L["Out of range"]], name)
			end
		end
	end
	if db.channel == "RAID" and mod.speaker and mod.speaker ~= UnitName("player") then
		print(TOKEN, ": skipping buff-check because ", mod.speaker, " is the speaker!")
		return
	end
	for k,v in pairs(buffStates) do
		broadCastToRaid(k, v)
	end
	if db.naggingWhispers then
		wipe(buffStates[L["Out of range"]])
		for k,v in pairs(buffStates) do
			broadCastWhisper(k, v)
		end
	end
	-- cleanup the mess
	wipe(buffStates)
	groups = nil
end

local dbDefaults = {
	profile = {
		suspend = false,
		naggingWhispers = false,
		threshold = 10*60, -- 10 minutes
		groups = { auto = true },
		channel = "SELF",
		flaskonly = false,
		checkfood = true,
		foodlevel = 0,
	}
}
for i = 1,8 do
	dbDefaults.profile.groups[i] = true
end

function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("BigBrother_ReadyDB", dbDefaults)
	db = self.db.profile
end

local ldbdata
function mod:OnEnable()
	debug("enabling")
	if db.suspend then
		ldbdata.icon = [[Interface\RAIDFRAME\ReadyCheck-NotReady]]
		ldbdata.text = "Off"
	else
		self:Resume()
	end
end

function mod:Resume()
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateMembers")
	self:RegisterEvent("READY_CHECK")
	self:RegisterEvent("READY_CHECK_FINISHED", "READY_CHECK")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	-- self:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT", "READY_CHECK")
	self:Resync()
	ldbdata.icon = [[Interface\RAIDFRAME\ReadyCheck-Ready]]
	if db.naggingWhispers then
		ldbdata.text = "On|cffff00ff/w|r"
	else
		ldbdata.text = "On"
	end
end

function mod:Suspend()
	self:UnregisterAllEvents()
	self:CancelAllTimers()
	self:Resync()
	ldbdata.icon = [[Interface\RAIDFRAME\ReadyCheck-NotReady]]
	ldbdata.text = "Off"
	mod.speaker = nil
end

function mod:OnDisable()
	self:Suspend()
end

local function toggleState()
	db.suspend = not db.suspend
	if db.suspend then
		mod:Suspend()
	else
		mod:Resume()
	end
end

ldbdata = {
	type = "launcher",
	OnClick = function(_, msg)
			if msg == "LeftButton" then
				if IsShiftKeyDown() then
					mod:READY_CHECK()
				else
					toggleState()
				end
			elseif msg == "RightButton" then
				mod:ShowConfig()
			end
		end,
	OnTooltipShow = function(tt)
			tt:AddLine("BigBrother - Ready")
			if not db.suspend then
				tt:AddLine("is |cff00ff00on|r")
			else
				tt:AddLine("is |cffff0000off|r")
			end
			if db.naggingWhispers then
				tt:AddLine("Additionally I will nag people with |cffff00ffwhipsers|r")
			end
			tt:AddLine("|cffffff00Click|r switch on/off")
			tt:AddLine("|cffffff00Shift-Click|r force buff-check")
			tt:AddLine("|cffffff00Right-Click|r open options")
			if mod.speaker then
				tt:AddLine("|cff00ff00 Speaker:|r " .. mod.speaker)
			end
		end,
	label = "BigBrother_Ready",
}

LibStub("LibDataBroker-1.1"):NewDataObject("BigBrother_Ready", ldbdata)
mod.ldbdata = ldbdata
-- sync-stuff
local candidates = {}
local ADDON_MSG_PREFIX = "BBReady"
RegisterAddonMessagePrefix(ADDON_MSG_PREFIX)
local playerName = UnitName("player")

function mod:Resync()
	debug("resyncing")
	SendAddonMessage(ADDON_MSG_PREFIX, "resync", "RAID")
end

local function monitorSpeaker(self)
	if not UnitIsConnected(mod.speaker) then
		debug("speaker went offline")
		mod:CheckGroup()
		mod:CancelTimer(timers.monitorSpeaker, true)
		timers.monitorSpeaker = nil
	end
end

function mod:ChooseCandidate()
	debug("timer expired. Choosing candidate now")
	if not next(candidates) then
		timers.ChooseCandidate = nil
		return
	end
	local bestvalue = 0
	local bestmembers = {}
	for k,v in pairs(candidates) do
		if v > bestvalue then
			bestvalue = v
			debug("resetting bestmembers-table")
			debug(k, v)
			wipe(bestmembers)
			table.insert(bestmembers,k)
		elseif v == bestvalue then
			debug("another one with same value")
			debug(k,v)
			table.insert(bestmembers,k)
		end
	end
	table.sort(bestmembers)
	mod.speaker = bestmembers[1]
	if(mod.speaker ~= playerName) then
		timers.monitorSpeaker = self:ScheduleRepeatingTimer(monitorSpeaker, 1)
	end
	debug(mod.speaker, " has been chosen")
	if mod.speaker ~= playerName then
		ldbdata.text = "|c88888800On|r"
	else
		ldbdata.text = "On"
	end
	wipe(candidates)
	timers.ChooseCandidate = nil
end

local function getVersionValue()
	local mversion, alpha = ("-"):split(mod.version)
	mversion = mversion:gsub("^v", "")
	alpha = tonumber(alpha) or 0
	local major, minor, rev = ("."):split(mversion)
	if not tonumber(major) then
		major = 99
	end
	return major * 1000000 + (minor or 0) * 10000 + (rev or 0) * 100 + alpha * 1
end

-- ping-value is calculated like: version > raidrank > naggingWhispers
local function getPingValue()
	local version = getVersionValue()
	local raidrank = (UnitIsGroupLeader("player") and 2) or (UnitIsGroupAssistant("player") and 1) or 0
	local nagValue = db.naggingWhispers and 1 or 0
	return version * 100 + raidrank * 10 + nagValue
end

local function sendPing()
	if db.channel == "SELF" then
		return
	end
	local message = getPingValue()
	debug("sending ping the raid with value", message)
	SendAddonMessage(ADDON_MSG_PREFIX, message, "RAID")
	timers.pingTimer = nil
end

local inRaid = false
function mod:UpdateMembers()
	local newState = not not UnitInRaid("player")
	debug(newState, "<->", inRaid)
	if newState ~= inRaid then
		debug("changed RAID-state")
		inRaid = newState
		self:Resync()
		return
	end
	if not UnitInRaid(mod.speaker) then
		debug("the SPEAKER left the raid")
		self:CheckGroup()
	end
end

function mod:CheckGroup(arg, ...)
	wipe(candidates)
	if timers.ChooseCandidate then
		debug("reschedule")
		self:CancelTimer(timers.ChooseCandidate, true)
	end
	debug("schedule timer")
	timers.ChooseCandidate = self:ScheduleTimer("ChooseCandidate", 10)
	self:CancelTimer(timers.sendping, true)
	timers.sendping = self:ScheduleTimer(sendPing, 2)
end

function mod:CHAT_MSG_ADDON(event, prefix, message, _, sender)
	--debug(prefix, " ", message, " ", sender)
	if prefix ~= ADDON_MSG_PREFIX then
		return
	end
	debug("message: ", message, " (", sender, ")")
	if message == "resync" then
		debug("recieved RESYNC from ", sender)
		self:CheckGroup()
		return
	end
	local rating = tonumber(message)
	if not rating then return end
	debug(sender, "comes with rating", rating)
	candidates[sender] = rating
end

function mod:PLAYER_REGEN_DISABLED()
	if not self:CheckForBoss() then
		return
	end
	if db.recheckAtBossFightStart then
		debug("checking again because of pull")
		self:READY_CHECK()
	end
	if db.repostAtBossFightEnd then
		self:CheckForWipe()
	end
end

function mod:CheckForBoss()
	if UnitExists("boss1") then
		debug("Engaged Boss: " .. tostring((UnitName("boss1"))))
		return true
	end
	return false
end

function mod:CheckForWipe()
	if self:DoCheckForWipe() then
		if db.repostAtBossFightEnd then
			print(TOKEN, " WIPED")
		end
	else
		self:ScheduleTimer("CheckForWipe", 1)
	end
end

function mod:DoCheckForWipe()
	local num = GetNumGroupMembers()
	for i = 1, num do
		local name = GetRaidRosterInfo(i)
		if name then
			if UnitAffectingCombat(name) ~= nil then
				return false
			end
		end
	end
	return true
end

local channelChoices = {
	RAID = "Raid",
	YELL = "Yell",
	SAY = "Say",
	SELF = "Self",
	-- CUSTOM = "Custom",
}
local function getChannel()
	return db.channel
end
local function setChannel(info, value)
	db.channel = value
	mod:Resync()
end

local groupChoices = {}
for i=1,8 do
	groupChoices[i] = i
end

local function getIncludeGroup(info, key)
	return db.groups[key]
end

local function setIncludeGroup(info, key, value)
	db.groups[key] = value
end

local function setOption(info, value)
	local key = info[#info]
	debug("setting ", key, " -> ", value)
	db[key] = value
end
local function getOption(info)
	local key = info[#info]
	return db[key]
end

local options = {
	name = "BigBrother - Ready",
	desc = "BigBrother - Ready",
	type = 'group',
	icon = [[Interface\GROUPFRAME\UI-Group-LeaderIcon]],
	args = {
		channel = {
			name = L["Channel"],
			desc = L["Which channel to broadcast to"],
			type = 'select',
			values = channelChoices,
			set = function(...)
				setOption(...)
				mod:Resync()
			end,
			get = getOption,
			order = 1,
		},
		naggingWhispers = {
			name = L["Nagging whispers"],
			desc = L["Additionally send whispers to players that are missing buffs"],
			type = 'toggle',
			get = getOption,
			set = function(info, value)
					db.naggingWhispers = value
					mod:Resync()
				end,
			order = 2,
		},
		threshold = {
			name = L["Expiration Threshold"],
			desc = L["Report Buffs that expire in x minutes"],
			type = 'range',
			min = 0,
			max = 60,
			step = 1,
			set = function(info, value) db.threshold = value*60 end,
			get = function(info) return db.threshold/60 end,
			order = 3,
		},
		flaskonly = {
			name = L["Flasks only"],
			desc = L["Only count flasks as valid buff, not elixirs"],
			type = 'toggle',
			get = getOption,
			set = setOption,
			order = 4,
		},
		checkfood = {
			name = L["Check food"],
			desc = L["Also check the players for food buffs"],
			type = 'toggle',
			get = getOption,
			set = setOption,
			order = 5,
		},
		foodlevel = {
			name = L["Food Level"],
			desc = L["Minimum stat bonus food buff must provide"],
			type = 'select',
			get = getOption,
			set = setOption,
			order = 6,
			values = {
				[0] = "Any",
				[250] = "250",
				[275] = "275",
				[300] = "300",
			},
		},
		recheckAtBossFightStart = {
			name = L["Check at boss"],
			desc = L["Check when a Boss is engaged"],
			type = 'toggle',
			get = getOption,
			set = setOption,
			order = 7,
		},
		includeGroups = {
			name = L["Groups"],
			desc = L["Only include these groups"],
			type = 'group',
			inline = true,
			args = {
				auto = {
					type = 'toggle',
					name = L["Auto"],
					desc = L["Determine Groups automatically"],
					get = function(info) return db.groups.auto end,
					set = function(info, value) db.groups.auto = value end,
				},
				includeGroups = {
					type = 'multiselect',
					name = L["Groups"],
					desc = L["Specifiy groups to include in the check"],
					values = groupChoices,
					get = getIncludeGroup,
					set = setIncludeGroup,
					disabled = function() return db.groups.auto end,
				},
			},
			order = 10,
		},
	},
}

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
AceConfig:RegisterOptionsTable("BBReady", options)
AceConfigDialog:SetDefaultSize("BBReady", 500, 350)
function mod:ShowConfig()
	AceConfigDialog:Open("BBReady")
end